@@ -11,10 +11,11 @@ import logging
 import warnings
 from bam_filter.utils import is_debug, calc_chunksize, fast_flatten, initializer
 import pyranges as pr
+from collections import defaultdict
 
 log = logging.getLogger("my_logger")
 
-sys.setrecursionlimit(10**6)
+sys.setrecursionlimit(10 ** 6)
 
 # Function to calculate evenness of coverage
 def coverage_evenness(coverage):
@@ -136,118 +137,90 @@ def get_bam_stats(params, ref_lengths=None, scale=1e6):
         ends.append(aln.reference_end)
         strands.append(strand)
 
-    if n_alns > 1:
-        # get bases covered by reads pileup
-        cov_pos = [
-            pileupcolumn.n
-            for pileupcolumn in samfile.pileup(
-                reference,
-                start=None,
-                stop=None,
-                region=None,
-                stepper="nofilter",
-                min_mapping_quality=0,
-            )
-        ]
-        # convert datafrane to pyranges
-        ranges = create_pyranges(reference, starts, ends, strands)
-        ranges = ranges.merge(strand=False).lengths().to_list()
-        max_covered_bases = np.max(ranges)
-        mean_covered_bases = np.mean(ranges)
-        bases_covered = int(len(cov_pos))
-        # get SD from covered bases
-        cov_sd = np.std(cov_pos, ddof=1)
-        # get average coverage
-        mean_coverage = sum(cov_pos) / reference_length
-        mean_coverage_covered = sum(cov_pos) / bases_covered
-
-        breadth = bases_covered / reference_length
-        exp_breadth = 1 - np.exp(-mean_coverage)
-        breadth_exp_ratio = breadth / exp_breadth
-
-        cov_evenness = coverage_evenness(cov_pos)
-        gc_content = (np.sum(read_gc_content) / np.sum(read_length)) * 100
-        c_v = cov_sd / mean_coverage
-        read_mapq = [np.nan if x == 255 else x for x in read_mapq]
-
-        tax_abund_aln = round((n_alns / reference_length) * scale)
-        tax_abund_read = round((len(set(read_names)) / reference_length) * scale)
-
-        log.debug(f"Number of reads: {len(set(read_names)):,}")
-        log.debug(f"Number of alignments: {n_alns:,}")
-        log.debug(f"Bases covered: {bases_covered:,}")
-        log.debug(f"Mean coverage: {mean_coverage:.2f}")
-        log.debug(f"Mean coverage covered: {mean_coverage_covered:.2f}")
-        log.debug(f"Max covered bases: {max_covered_bases:,}")
-        log.debug(f"Mean covered bases: {mean_covered_bases:.2f}")
-        log.debug(f"SD: {cov_sd:.2f}")
-        log.debug(f"Breadth: {breadth:.2f}")
-        log.debug(f"Exp. breadth: {exp_breadth:.2f}")
-        log.debug(f"Breadth/exp. ratio: {breadth_exp_ratio:.2f}")
-        log.debug(f"Cov. evenness: {cov_evenness:.2f}")
-        log.debug(f"C_v: {c_v:.2f}")
-        log.debug(f"Mean mapq: {np.mean(read_mapq):.2f}")
-        log.debug(f"GC content: {gc_content:.2f}")
-        log.debug(f"Taxonomic abundance (alns): {tax_abund_aln:.2f}")
-        log.debug(f"Taxonomic abundance (reads): {tax_abund_read:.2f}")
-        data = BamAlignment(
-            reference=reference,
-            n_alns=n_alns,
-            reference_length=reference_length,
-            bam_reference_length=bam_reference_length,
-            mean_coverage=mean_coverage,
-            mean_coverage_covered=mean_coverage_covered,
-            bases_covered=bases_covered,
-            max_covered_bases=max_covered_bases,
-            mean_covered_bases=mean_covered_bases,
-            cov_evenness=cov_evenness,
-            breadth=breadth,
-            exp_breadth=exp_breadth,
-            breadth_exp_ratio=breadth_exp_ratio,
-            c_v=c_v,
-            edit_distances=edit_distances,
-            # edit_distances_md=edit_distances_md,
-            ani_nm=ani_nm,
-            # ani_md=ani_md,
-            read_length=read_length,
-            read_gc_content=read_gc_content,
-            read_aligned_length=read_aligned_length,
-            mapping_quality=read_mapq,
-            read_names=set(read_names),
-            read_aln_score=read_aln_score,
-            tax_abund_aln=tax_abund_aln,
-            tax_abund_read=tax_abund_read,
+    # get bases covered by reads pileup
+    cov_pos = [
+        pileupcolumn.n
+        for pileupcolumn in samfile.pileup(
+            reference,
+            start=None,
+            stop=None,
+            region=None,
+            stepper="nofilter",
+            min_mapping_quality=0,
         )
-    else:
-        # data = BamAlignment(
-        #     reference=reference,
-        #     n_alns=n_alns,
-        #     reference_length=reference_length,
-        #     bam_reference_length=bam_reference_length,
-        #     mean_coverage=np.nan,
-        #     mean_coverage_covered=np.nan,
-        #     bases_covered=np.nan,
-        #     max_covered_bases=np.nan,
-        #     mean_covered_bases=np.nan,
-        #     cov_evenness=np.nan,
-        #     breadth=np.nan,
-        #     exp_breadth=np.nan,
-        #     breadth_exp_ratio=np.nan,
-        #     c_v=np.nan,
-        #     edit_distances=np.nan,
-        #     # edit_distances_md=np.nan,
-        #     ani_nm=np.nan,
-        #     # ani_md=np.nan,
-        #     read_length=np.nan,
-        #     read_gc_content=np.nan,
-        #     read_aligned_length=np.nan,
-        #     mapping_quality=np.nan,
-        #     read_names=read_names,
-        #     read_aln_score=read_aln_score,
-        #     tax_abund_aln=np.nan,
-        #     tax_abund_read=np.nan,
-        # )
-        data = None
+    ]
+    # convert datafrane to pyranges
+    ranges = create_pyranges(reference, starts, ends, strands)
+    ranges = ranges.merge(strand=False).lengths().to_list()
+    max_covered_bases = np.max(ranges)
+    mean_covered_bases = np.mean(ranges)
+    bases_covered = int(len(cov_pos))
+    # get SD from covered bases
+    cov_sd = np.std(cov_pos, ddof=1)
+    # get average coverage
+    mean_coverage = sum(cov_pos) / reference_length
+    mean_coverage_covered = sum(cov_pos) / bases_covered
+
+    breadth = bases_covered / reference_length
+    exp_breadth = 1 - np.exp(-mean_coverage)
+    breadth_exp_ratio = breadth / exp_breadth
+
+    if breadth_exp_ratio > 1:
+        breadth_exp_ratio = 1.0
+
+    cov_evenness = coverage_evenness(cov_pos)
+    gc_content = (np.sum(read_gc_content) / np.sum(read_length)) * 100
+    c_v = cov_sd / mean_coverage
+    read_mapq = [np.nan if x == 255 else x for x in read_mapq]
+
+    tax_abund_aln = round((n_alns / reference_length) * scale)
+    tax_abund_read = round((len(set(read_names)) / reference_length) * scale)
+
+    log.debug(f"Number of reads: {len(set(read_names)):,}")
+    log.debug(f"Number of alignments: {n_alns:,}")
+    log.debug(f"Bases covered: {bases_covered:,}")
+    log.debug(f"Mean coverage: {mean_coverage:.2f}")
+    log.debug(f"Mean coverage covered: {mean_coverage_covered:.2f}")
+    log.debug(f"Max covered bases: {max_covered_bases:,}")
+    log.debug(f"Mean covered bases: {mean_covered_bases:.2f}")
+    log.debug(f"SD: {cov_sd:.2f}")
+    log.debug(f"Breadth: {breadth:.2f}")
+    log.debug(f"Exp. breadth: {exp_breadth:.2f}")
+    log.debug(f"Breadth/exp. ratio: {breadth_exp_ratio:.2f}")
+    log.debug(f"Cov. evenness: {cov_evenness:.2f}")
+    log.debug(f"C_v: {c_v:.2f}")
+    log.debug(f"Mean mapq: {np.mean(read_mapq):.2f}")
+    log.debug(f"GC content: {gc_content:.2f}")
+    log.debug(f"Taxonomic abundance (alns): {tax_abund_aln:.2f}")
+    log.debug(f"Taxonomic abundance (reads): {tax_abund_read:.2f}")
+    data = BamAlignment(
+        reference=reference,
+        n_alns=n_alns,
+        reference_length=reference_length,
+        bam_reference_length=bam_reference_length,
+        mean_coverage=mean_coverage,
+        mean_coverage_covered=mean_coverage_covered,
+        bases_covered=bases_covered,
+        max_covered_bases=max_covered_bases,
+        mean_covered_bases=mean_covered_bases,
+        cov_evenness=cov_evenness,
+        breadth=breadth,
+        exp_breadth=exp_breadth,
+        breadth_exp_ratio=breadth_exp_ratio,
+        c_v=c_v,
+        edit_distances=edit_distances,
+        # edit_distances_md=edit_distances_md,
+        ani_nm=ani_nm,
+        # ani_md=ani_md,
+        read_length=read_length,
+        read_gc_content=read_gc_content,
+        read_aligned_length=read_aligned_length,
+        mapping_quality=read_mapq,
+        read_names=set(read_names),
+        read_aln_score=read_aln_score,
+        tax_abund_aln=tax_abund_aln,
+        tax_abund_read=tax_abund_read,
+    )
     return data
 
 
@@ -357,7 +330,7 @@ class BamAlignment:
             edit_distances = np.mean(self.edit_distances)
             read_ani_mean = np.mean(self.ani_nm)
             read_ani_std = np.std(self.ani_nm, ddof=1)
-            reads_ani_median = np.median(self.ani_nm)
+            read_ani_median = np.median(self.ani_nm)
             gc_content = (np.sum(self.read_gc_content) / np.sum(self.read_length)) * 100
         return {
             "reference": self.reference,
@@ -377,7 +350,7 @@ class BamAlignment:
             # "edit_distances_md": np.mean(self.edit_distances_md),
             "read_ani_mean": read_ani_mean,
             "read_ani_std": read_ani_std,
-            "reads_ani_median": reads_ani_median,
+            "read_ani_median": read_ani_median,
             # "ani_md": np.mean(self.ani_md),
             "bases_covered": self.bases_covered,
             "max_covered_bases": self.max_covered_bases,
@@ -452,6 +425,25 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6):
     total_refs = samfile.nreferences
     logging.info(f"Found {total_refs:,} reference sequences")
     # logging.info(f"Found {samfile.mapped:,} alignments")
+    logging.info(f"Removing references without mappings...")
+    # Remove references without mapped reads
+    alns_in_ref = defaultdict(int)
+    for aln in tqdm.tqdm(
+        samfile.fetch(until_eof=True),
+        total=samfile.mapped,
+        leave=False,
+        ncols=80,
+        desc=f"Alignments processed",
+    ):
+        alns_in_ref[aln.reference_name] += 1
+
+    references = [ref for ref, count in alns_in_ref.items() if count > 0]
+
+    if len(references) == 0:
+        logging.error("No reference sequences with alignments found in the BAM file")
+        sys.exit(1)
+
+    logging.info(f"Keeping {len(references):,} references")
 
     params = zip([bam] * len(references), references)
     try:
@@ -474,8 +466,7 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6):
                 initargs=([params, ref_lengths, scale],),
             )
             c_size = calc_chunksize(
-                n_workers=threads,
-                len_iterable=len(references),
+                n_workers=threads, len_iterable=len(references), factor=4
             )
             data = list(
                 tqdm.tqdm(
@@ -484,7 +475,7 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6):
                             get_bam_stats, ref_lengths=ref_lengths, scale=scale
                         ),
                         params,
-                        chunksize=c_size,
+                        chunksize=20,
                     ),
                     total=len(references),
                     leave=False,
@@ -500,11 +491,20 @@ def process_bam(bam, threads=1, reference_lengths=None, scale=1e6):
         logging.info(f"User canceled the operation. Terminating jobs.")
         p.terminate()
         p.join()
-        sys.exit()
+        sys.exit(0)
     return data
 
 
-def filter_reference_BAM(bam, df, filter_conditions, threads, out_files, sort_memory):
+def filter_reference_BAM(
+    bam,
+    df,
+    filter_conditions,
+    threads,
+    out_files,
+    sort_memory,
+    only_stats_filtered,
+    sort_by_name=False,
+):
     """Filter BAM based on certain conditions
 
     Args:
@@ -515,95 +515,113 @@ def filter_reference_BAM(bam, df, filter_conditions, threads, out_files, sort_me
     """
     logging.info(f"Filtering stats...")
     logging.info(
-        f"min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_read_ani >= {filter_conditions['min_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']}"
+        f"min_read_count >= {filter_conditions['min_read_count']} & min_read_length >= {filter_conditions['min_read_length']} & min_read_ani >= {filter_conditions['min_read_ani']} & min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} &  min_breadth >= {filter_conditions['min_breadth']} & min_coverage_evenness >= {filter_conditions['min_coverage_evenness']}"
     )
-
     df_filtered = df.loc[
         (df["n_reads"] >= filter_conditions["min_read_count"])
         & (df["read_length_mean"] >= filter_conditions["min_read_length"])
         & (df["read_ani_mean"] >= filter_conditions["min_read_ani"])
         & (df["breadth_exp_ratio"] >= filter_conditions["min_expected_breadth_ratio"])
+        & (df["breadth"] >= filter_conditions["min_breadth"])
         & (df["cov_evenness"] >= filter_conditions["min_coverage_evenness"])
     ]
     if len(df_filtered.index) > 0:
-        refs_dict = dict(zip(df_filtered["reference"], df_filtered["reference_length"]))
-        (ref_names, ref_lengths) = zip(*refs_dict.items())
-
-        out_bam_file = pysam.Samfile(
-            out_files["bam_filtered_tmp"],
-            "wb",
-            referencenames=list(ref_names),
-            referencelengths=list(ref_lengths),
-            threads=threads,
-        )
-        header = pysam.AlignmentHeader.from_references(
-            list(ref_names), list(ref_lengths)
-        )
-        logging.info(f"Writing filtered BAM file... (be patient)")
-        references = df_filtered["reference"].values
-        params = zip([bam] * len(references), references)
         logging.info(f"Saving filtered stats...")
         df_filtered.to_csv(
             out_files["stats_filtered"], sep="\t", index=False, compression="gzip"
         )
-        try:
-            logging.info(f"Filtering BAM file...")
+        if only_stats_filtered:
+            logging.info(f"Skipping saving filtered BAM file.")
+        else:
+            logging.info(f"Writing filtered BAM file... (be patient)")
+            refs_dict = dict(
+                zip(df_filtered["reference"], df_filtered["reference_length"])
+            )
+            (ref_names, ref_lengths) = zip(*refs_dict.items())
 
-            if is_debug():
-                alns = list(map(get_alns, params))
-            else:
+            out_bam_file = pysam.Samfile(
+                out_files["bam_filtered_tmp"],
+                "wb",
+                referencenames=list(ref_names),
+                referencelengths=list(ref_lengths),
+                threads=threads,
+            )
+            header = pysam.AlignmentHeader.from_references(
+                list(ref_names), list(ref_lengths)
+            )
+            references = df_filtered["reference"].values
+            params = zip([bam] * len(references), references)
+            try:
+                logging.info(f"Filtering BAM file...")
 
-                p = Pool(threads)
-                c_size = calc_chunksize(threads, len(references))
-                alns = list(
-                    tqdm.tqdm(
-                        p.imap_unordered(get_alns, params, chunksize=c_size),
-                        total=len(references),
-                        leave=False,
-                        ncols=80,
-                        desc=f"References processed",
+                if is_debug():
+                    alns = list(map(get_alns, params))
+                else:
+
+                    p = Pool(threads)
+                    c_size = calc_chunksize(threads, len(references))
+                    alns = list(
+                        tqdm.tqdm(
+                            p.imap_unordered(get_alns, params, chunksize=c_size),
+                            total=len(references),
+                            leave=False,
+                            ncols=80,
+                            desc=f"References processed",
+                        )
                     )
+
+                    p.close()
+                    p.join()
+
+            except KeyboardInterrupt:
+                logging.info(f"User canceled the operation. Terminating jobs.")
+                p.terminate()
+                p.join()
+                sys.exit(0)
+
+            for aln in fast_flatten(alns):
+                out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header=header))
+            out_bam_file.close()
+            if sort_by_name:
+                logging.info(f"Sorting BAM file by read name...")
+                pysam.sort(
+                    "-n",
+                    "-@",
+                    str(threads),
+                    "-m",
+                    str(sort_memory),
+                    "-o",
+                    out_files["bam_filtered"],
+                    out_files["bam_filtered_tmp"],
+                )
+            else:
+                pysam.sort(
+                    "-@",
+                    str(threads),
+                    "-m",
+                    str(sort_memory),
+                    "-o",
+                    out_files["bam_filtered"],
+                    out_files["bam_filtered_tmp"],
                 )
 
-                p.close()
-                p.join()
+                save = pysam.set_verbosity(0)
+                samfile = pysam.AlignmentFile(out_files["bam_filtered"], "rb")
+                chr_lengths = []
+                for chrom in samfile.references:
+                    chr_lengths.append(samfile.get_reference_length(chrom))
+                max_chr_length = np.max(chr_lengths)
+                pysam.set_verbosity(save)
+                samfile.close()
 
-        except KeyboardInterrupt:
-            logging.info(f"User canceled the operation. Terminating jobs.")
-            p.terminate()
-            p.join()
-            sys.exit()
+                logging.info(f"BAM index not found. Indexing...")
+                if max_chr_length > 536870912:
+                    logging.info(f"A reference is longer than 2^29, indexing with csi")
+                    pysam.index(out_files["bam_filtered"], "-c")
+                else:
+                    pysam.index(out_files["bam_filtered"])
 
-        for aln in fast_flatten(alns):
-            out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header=header))
-        out_bam_file.close()
-        pysam.sort(
-            "-@",
-            str(threads),
-            "-m",
-            str(sort_memory),
-            "-o",
-            out_files["bam_filtered"],
-            out_files["bam_filtered_tmp"],
-        )
-
-        save = pysam.set_verbosity(0)
-        samfile = pysam.AlignmentFile(out_files["bam_filtered"], "rb")
-        chr_lengths = []
-        for chrom in samfile.references:
-            chr_lengths.append(samfile.get_reference_length(chrom))
-        max_chr_length = np.max(chr_lengths)
-        pysam.set_verbosity(save)
-        samfile.close()
-
-        logging.info(f"BAM index not found. Indexing...")
-        if max_chr_length > 536870912:
-            logging.info(f"A reference is longer than 2^29, indexing with csi")
-            pysam.index(out_files["bam_filtered"], "-c")
-        else:
-            pysam.index(out_files["bam_filtered"])
-
-        os.remove(out_files["bam_filtered_tmp"])
+            os.remove(out_files["bam_filtered_tmp"])
     else:
         logging.info(f"No references meet the filter conditions. Skipping...")
 
